/*
 * Dual MAX31865/PT100 precision temperature monitor
 *   - ESP32 WROOM-based dev board
 *   - Use two Adafruit MAX31865 breakout boards with 4-wire PT100 temp sensor
 *   - 2x8 line 4-bit parallel LCD display
 *   - Simple dual serial/Bluetooth serial interface for external temperature logging
 * 
 * Connections
 *     0 - Button (active low with built-in pull-up)
 *    12 - MAX31865 MISO
 *    13 - MAX31865 MOSI
 *    14 - MAX31865 SCLK
 *    15 - Channel 1 MAX31865 CSN
 *    16 - Channel 2 MAX31865 CSN
 *    17 - LCD RS
 *    18 - LCD EN
 *    19 - LCD D4
 *    21 - LCD D5
 *    22 - LCD D6
 *    23 - LCD D7
 *    
 *    LCD RW tied low
 */
#include <LiquidCrystal.h>
#include <Adafruit_MAX31865.h>
#include "BluetoothSerial.h"

// =============================
// Constants
//

#define NUM_TEMPS        2
#define VERSION_STRING   "1.0"

// Sample update rate
#define UPDATE_MSEC      500

// Display update rate (should be slower than or equal to UPDATE_MSEC)
#define DISP_UPDATE_MSEC 1000

// Number of display samples
#define DISP_AVG_COUNT   4

// The value of the Rref resistor (ohms)
#define RREF 430.0

// Command processor
#define CMD_ST_IDLE 0
#define CMD_ST_CMD  1
#define CMD_ST_ARG  2

#define TERM_CHAR 0x0D


// =============================
// Hardware pins
//
const int pin_btn    = 0;

const int pin_mosi   = 13;
const int pin_miso   = 12;
const int pin_sclk   = 14;
const int pin_t1_csn = 15;
const int pin_t2_csn = 16;

const int pin_lcd_rs = 17;
const int pin_lcd_en = 18;
const int pin_lcd_d4 = 19;
const int pin_lcd_d5 = 21;
const int pin_lcd_d6 = 22;
const int pin_lcd_d7 = 23;



// =============================
// Hardware objects
//
Adafruit_MAX31865 max_t[NUM_TEMPS] = {
   Adafruit_MAX31865(pin_t1_csn, pin_mosi, pin_miso, pin_sclk),
   Adafruit_MAX31865(pin_t2_csn, pin_mosi, pin_miso, pin_sclk)
};

BluetoothSerial SerialBT;

LiquidCrystal lcd(pin_lcd_rs, pin_lcd_en, pin_lcd_d4, pin_lcd_d5, pin_lcd_d6, pin_lcd_d7);



// =============================
// Variables
//

// Control state
bool log_enable[2];
int disp_update_count;
int log_count;
int cur_log_count;
unsigned long prevT;
unsigned long zeroT;
uint8_t max_status[NUM_TEMPS];

// Temperature sampling
float cur_temp[NUM_TEMPS];

// Temperature averaging for display
float avg_buf[NUM_TEMPS][DISP_AVG_COUNT];
float avg_temp[NUM_TEMPS];
int avg_push_index[NUM_TEMPS];

// Command processor
int CmdState[2];
char CurCmd[2];
uint32_t CmdIndex[2];
uint32_t CmdArg[2];
boolean CmdHasArg[2];

// Help Contents
static const char* help_string = \
"Command Interface:\n\r" \
"  H       : Help\n\r" \
"  L/L=<N> : Log Interval (mSec - multiple of 500 - 0 to disable)\n\r" \
"  T<I>    : Temperature (I = channel 0..N)\n\r" \
"  S<I>    : Status (I = channel 0..N)\n\r" \
"  V       : Version\n\r" \
"  Z       : Zero Timestamp\n\r";



// =============================
// Arduino Entry points
//
void setup() {
  int i;

  // Setup hardware
  Serial.begin(115200);
  Serial.println("Dual channel temperature monitor - Version " VERSION_STRING);

  SerialBT.begin("TempMon");
  
  lcd.begin(8, 2);
  lcd.clear();
  lcd.setCursor(0, 0);
  lcd.print("TempMon");
  lcd.setCursor(0, 1);
  lcd.print("V " VERSION_STRING);

  for (i=0; i<NUM_TEMPS; i++) {
    max_t[i].begin(MAX31865_4WIRE);
    avg_push_index[i] = 0;
  }

  pinMode(pin_btn, INPUT_PULLUP);

  // Get an initial temperature average
  for (i=0; i<DISP_AVG_COUNT; i++) {
    UpdateTemps();
  }

  // Setup for operation
  log_enable[0] = false;
  log_enable[1] = false;
  log_count = 0;
  cur_log_count = 0;
  disp_update_count = 0;
  prevT = millis();
  zeroT = prevT;
}


void loop() {
  char c;
  
  // Evaluate temperature sensors
  if (InactivityTimeout(UPDATE_MSEC)) {
    UpdateTemps();

    if (++disp_update_count == (DISP_UPDATE_MSEC/UPDATE_MSEC)) {
      disp_update_count = 0;
      UpdateDisplay();
    }

    if (log_enable[0] || log_enable[1]) {
      if (--cur_log_count == 0) {
        cur_log_count = log_count;
        if (log_enable[0]) {
          LogTemps(0);
        }
        if (log_enable[1]) {
          LogTemps(1);
        }
      }
    }
  }

  // Evaluate the serial enable
  if (Serial.available()) {
    c = Serial.read();
    EvalCommand(c, 0);
  }
  
  if (SerialBT.available()) {
    c = SerialBT.read();
    EvalCommand(c, 1);
  }

  if ((CmdState[0] == CMD_ST_IDLE) && (CmdState[1] == CMD_ST_IDLE)) {
    // Delay to allow FreeRTOS task scheduling while idle
    delay(20);
  }
}



// =============================
// Subroutines
//

void UpdateTemps()
{
  int i, j;
  float sum;

  for (i=0; i<NUM_TEMPS; i++) {
    max_status[i] = max_t[i].readFault();
    if (max_status[i] == 0) {
      // Push the current temperatures
      cur_temp[i] = max_t[i].temperature(100, RREF);
      avg_buf[i][avg_push_index[i]] = cur_temp[i];
      avg_push_index[i] = avg_push_index[i] + 1;
      if (avg_push_index[i] == DISP_AVG_COUNT) {
        avg_push_index[i] = 0;
      }

      // Update the average temperature value
      sum = 0;
      for (j=0; j<DISP_AVG_COUNT; j++) {
        sum += avg_buf[i][j];
      }
      avg_temp[i] = sum / DISP_AVG_COUNT;
    } else {
      max_t[i].clearFault();
      
      Serial.printf("Sensor %d Fault 0x%2x:\n", i, max_status[i]);
      if (max_status[i] & MAX31865_FAULT_HIGHTHRESH) {
        Serial.println("  RTD High Threshold"); 
      }
      if (max_status[i] & MAX31865_FAULT_LOWTHRESH) {
        Serial.println("  RTD Low Threshold"); 
      }
      if (max_status[i] & MAX31865_FAULT_REFINLOW) {
        Serial.println("  REFIN- > 0.85 x Bias"); 
      }
      if (max_status[i] & MAX31865_FAULT_REFINHIGH) {
        Serial.println("  REFIN- < 0.85 x Bias - FORCE- open"); 
      }
      if (max_status[i] & MAX31865_FAULT_RTDINLOW) {
        Serial.println("  RTDIN- < 0.85 x Bias - FORCE- open"); 
      }
      if (max_status[i] & MAX31865_FAULT_OVUV) {
        Serial.println("  Under/Over voltage"); 
      }
    }
  }
}


void UpdateDisplay()
{
  // Top Line
  lcd.setCursor(0, 0);
  if (max_status[0] == 0) {
    lcd.print("< ");
    lcd.print(avg_temp[0]);
    lcd.print("C");
  } else {
    lcd.print("<FLT: ");
    lcd.print(max_status[0], HEX);
  }

  // Bottom Line
  lcd.setCursor(0, 1);
  if (max_status[1] == 0) {
    lcd.print(avg_temp[1]);
    lcd.print("C >");
  } else {
    lcd.print("FLT: ");
    lcd.print(max_status[1], HEX);
    lcd.print(">");
  }
}


void LogTemps(int src)
{
  float secs;
  int i;

  secs = (prevT - zeroT) / 1000.0;
  
  if (src == 0) {
    Serial.print(secs);
    Serial.print(" ");
  } else {
    SerialBT.print(secs);
    SerialBT.print(" ");
  }
  
  for (i=0; i<NUM_TEMPS; i++) {
    if (src == 0) {
      Serial.print(cur_temp[i]);
      Serial.print(" ");
    } else {
      SerialBT.print(cur_temp[i]);
      SerialBT.print(" ");
    }
  }

  if (src == 0) {
    Serial.println();
  } else {
    SerialBT.println();
  }
}


boolean InactivityTimeout(unsigned long timeout) {
  unsigned long curT = millis();
  unsigned long deltaT;
  
  if (curT > prevT) {
    deltaT = curT - prevT;
  } else {
    // Handle wrap
    deltaT = ~(prevT - curT) + 1;
  }
  
  if (deltaT >= timeout) {
    prevT = curT;
    return true;
  }
  
  return false;
}
